/* src/Osi/config_osi.h.  Generated by configure.  */
/* src/Osi/config_osi.h.in. */

#ifndef __CONFIG_OSI_H__
#define __CONFIG_OSI_H__

/* Version number of project */
#define OSI_VERSION "0.108.7"

/* Major Version number of project */
#define OSI_VERSION_MAJOR 0

/* Minor Version number of project */
#define OSI_VERSION_MINOR 108

/* Release Version number of project */
#define OSI_VERSION_RELEASE 7

#endif
